/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.bc.sql.listener;

import java.io.InputStream;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.concurrent.ExecutorService;
import java.util.logging.Level;

import javax.jbi.messaging.MessagingException;
import javax.sql.DataSource;
import javax.xml.namespace.QName;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.bc.sql.Constants;
import org.ow2.petals.bc.sql.DataSources;
import org.ow2.petals.bc.sql.SqlComponent;
import org.ow2.petals.bc.sql.service.SQLService;
import org.ow2.petals.component.framework.api.Message;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.component.framework.util.XMLUtil;
import org.w3c.dom.Element;

/**
 * @author alouis
 * 
 */
public class JBIListener extends AbstractJBIListener {

    private DataSources dataSources;

    private SQLService sqlService;

    private ExecutorService threadPool;

    /**
     * Creates a new instance of {@link JBIListener}
     * 
     */
    public JBIListener() {
    }

    @Override
    public void init() {
        this.dataSources = ((SqlComponent) this.getComponent()).getDataSources();
        this.sqlService = ((SqlComponent) this.getComponent()).getSQLService();
        this.threadPool = ((SqlComponent) this.getComponent()).getThreadPool();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.listener.AbstractJBIListener#onJBIMessage
     * (org.objectweb.petals.component.framework.util.Exchange)
     */
    @Override
    public final boolean onJBIMessage(final Exchange exchange) {
        Connection connection = null;

        try {
            if (!validateExchange(exchange)) {
                this.getLogger().log(Level.WARNING, exchange.getError().getMessage());
                // no SQL request, error is set in the exchange
                return true;
            }

            final String sql = getSQLRequest(exchange);
            if (exchange.getError() != null) {
                this.getLogger().log(Level.WARNING, exchange.getError().getMessage());
                // no SQL request, error is set in the exchange
                return true;
            }

            // get a JDBC connection
            final DataSource dataSource = this.dataSources.getDataSource(this.getProvides());
            if (dataSource == null) {
                throw new Exception("No dataSource has been found for the service "
                        + exchange.getService());
            }
            // open the connection
            this.getLogger().finest("Get a JDBC connection from the DataSource");
            connection = dataSource.getConnection();

            final String operation = exchange.getOperationName().toLowerCase();
            if (Constants.OPERATION_SELECT.equals(operation)) {
                final ResultSet resultSet = this.sqlService.select(sql, connection);

                final boolean showMetadata = this.getShowMetadata(exchange);
                if (showMetadata) {
                    this.getLogger().finest("Metadata written in the response");
                }
                // response is written in a separate thread, while the response
                // is sent back in the current one
                final OutStreamer outStreamer = new OutStreamer(this.sqlService, resultSet,
                        showMetadata, connection, this.getLogger());
                this.threadPool.execute(outStreamer);
                final InputStream inputStream = outStreamer.getPipedInputStream();
                exchange.setOutMessageContent(new StreamSource(inputStream));
                connection = null;
            } else if (Constants.OPERATION_INSERT.equals(operation)
                    || Constants.OPERATION_UPDATE.equals(operation)
                    || Constants.OPERATION_DELETE.equals(operation)) {
                this.sqlService.update(sql, connection);
            } else {
                final int result = this.sqlService.update(sql, connection);
                final String resultXML = "<updated xmlns=\"http://petals.ow2.org/components/sql/version-1\">"
                    + result + "</updated>";
                exchange.getOutMessage().setContent(
                        UtilFactory.getSourceUtil().createSource(resultXML));
            }

        } catch (final Exception e) {
            this.getLogger().log(Level.SEVERE, e.getMessage(), e);
            exchange.setError(new Exception(e));
        } finally {
            // close the connection (except for streaming mode)
            if (connection != null) {
                try {
                    connection.close();
                } catch (final SQLException e) {
                    this.getLogger().log(Level.WARNING, e.getMessage(), e);
                }
            }
        }

        return true;
    }

    /**
     * Get the SQL request from the exchange. If failed, an error is set to the
     * exchange.
     * 
     * @param document
     * @return
     * @throws MessagingException
     */
    private static final String getSQLRequest(final Exchange exchange) throws MessagingException {
        String sql = null;
        final Element queryElement = exchange.getInMessageContentAsDocument().getDocumentElement();
        if (!(Constants.SQL_ELEMENT.equals(queryElement.getLocalName()) && Constants.NAMESPACE
                .equals(queryElement.getNamespaceURI()))) {
            exchange.setError(new Exception(
                    "The request message must contain a root element named '"
                    + Constants.SQL_ELEMENT + "' in the namespace '" + Constants.NAMESPACE
                    + "'"));
        } else {
            sql = XMLUtil.getTextContent(queryElement);
        }

        return sql;
    }

    /**
     * Is Show Metadata activated?
     * 
     * @param exchange
     * @return
     */
    private final boolean getShowMetadata(final Exchange exchange) {
        boolean showMetadata = false;
        if (exchange.getProperty(Constants.METADATA) != null
                && "true".equals(exchange.getProperty(Constants.METADATA))) {
            showMetadata = true;
        } else if (this.getExtensions().get(Constants.METADATA, "false").equals("true")) {
            showMetadata = true;
        }
        try {
            if (exchange.getInMessageProperty(Constants.METADATA) != null
                    && "true".equals(exchange.getInMessageProperty(Constants.METADATA)) ) {
                showMetadata = true;
            }
        } catch (final MessagingException e) {
            //do nothing
        }


        return showMetadata;
    }

    /**
     * Validate the operation and the MEP.
     * 
     * @param operation
     */
    private static final boolean validateExchange(final Exchange exchange) {
        boolean result = true;

        final QName operation = exchange.getOperation();
        if (!operation.getNamespaceURI().equals(Constants.NAMESPACE)) {
            exchange.setError(new Exception("Operation '" + operation
                    + "' must be in the namespace '" + Constants.NAMESPACE + "'"));
            result = false;
        }

        if (Constants.OPERATION_SELECT.equals(operation.getLocalPart().toLowerCase())) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new Exception("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_INSERT.equals(operation.getLocalPart().toLowerCase())) {
            if (!Message.MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new Exception("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_ONLY_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_INSERT_WITH_RESPONSE.equals(operation.getLocalPart()
                .toLowerCase())) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new Exception("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_UPDATE.equals(operation.getLocalPart().toLowerCase())) {
            if (!Message.MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new Exception("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_ONLY_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_UPDATE_WITH_RESPONSE.equals(operation.getLocalPart()
                .toLowerCase())) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new Exception("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_DELETE.equals(operation.getLocalPart().toLowerCase())) {
            if (!Message.MEPConstants.IN_ONLY_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new Exception("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_ONLY_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else if (Constants.OPERATION_DELETE_WITH_RESPONSE.equals(operation.getLocalPart()
                .toLowerCase())) {
            if (!Message.MEPConstants.IN_OUT_PATTERN.equals(exchange.getPattern())) {
                exchange.setError(new Exception("Operation '" + operation
                        + "' must use the pattern '" + Message.MEPConstants.IN_OUT_PATTERN.value()
                        + "'"));
                result = false;
            }
        } else {
            exchange.setError(new Exception("Operation '" + operation
                    + "' must have a declared local name. Choices are :'"
                    + Constants.OPERATION_SELECT + "','" + Constants.OPERATION_INSERT + "','"
                    + Constants.OPERATION_INSERT_WITH_RESPONSE + "','" + Constants.OPERATION_UPDATE
                    + Constants.OPERATION_UPDATE_WITH_RESPONSE + "','" + "','"
                    + Constants.OPERATION_DELETE + "'" + Constants.OPERATION_DELETE_WITH_RESPONSE));
        }

        return result;
    }
}
